@@ -56,9 +56,7 @@ class PFEnergySeriesIntegrator(PFGeneralIntegrator):
             return res
     def integrateImageStack(self,img_stack):
         
-        if img_stack.system.to_pandas().drop_duplicates().shape[0] != img_stack.system.shape[0]:
-            warnings.warn('Your system contains duplicate conditions.  This is not supported and may not work.  Try adding additional coords to separate image conditions',stacklevel=2)
-        
+       
         # get just the energies of the image stack
         energies = img_stack.energy.to_dataframe()
         
@@ -83,9 +81,27 @@ class PFEnergySeriesIntegrator(PFGeneralIntegrator):
         # single image reduce each entry in the stack
         # + 
         # restack the reduced data
-
-        return img_stack.groupby('system',squeeze=False).progress_apply(self.integrateSingleImage)
+        data = img_stack
+        indexes = list(data.indexes.keys())
+        indexes.remove('pix_x')
+        indexes.remove('pix_y')
+        
+        if len(indexes) == 1:
+            if img_stack.__getattr__(indexes[0]).to_pandas().drop_duplicates().shape[0] != img_stack.__getattr__(indexes[0]).shape[0]:
+                warnings.warn(f'Axis {indexes[0]} contains duplicate conditions.  This is not supported and may not work.  Try adding additional coords to separate image conditions',stacklevel=2)
+            data_int = data.groupby(indexes[0],squeeze=False).progress_apply(self.integrateSingleImage)
+        else:
+            #some kinda logic to check for existing multiindexes and stack into them appropriately maybe
+            data = data.stack({'pyhyper_internal_multiindex':indexes})
+            if img_stack.pyhyper_internal_multiindex.to_pandas().drop_duplicates().shape[0] != img_stack.pyhyper_internal_multiindex.shape[0]:
+                warnings.warn('Your index set contains duplicate conditions.  This is not supported and may not work.  Try adding additional coords to separate image conditions',stacklevel=2)
+        
+            data_int = data.groupby('pyhyper_internal_multiindex',squeeze=False).progress_apply(self.integrateSingleImage).unstack('pyhyper_internal_multiindex')
+        return data_int
+        #return img_stack.groupby('system',squeeze=False).progress_apply(self.integrateSingleImage)
     
+
+
     def createIntegrator(self,en):
         self.integrator_stack[en] = azimuthalIntegrator.AzimuthalIntegrator(
             self.dist, self.poni1, self.poni2, self.rot1, self.rot2, self.rot3 ,pixel1=self.pixel1,pixel2=self.pixel2, wavelength = 1.239842e-6/en)
